import logging

from rich.table import Table
from rich import box
from rich.console import Console

from .schema import Schema
from .types import BuildError, ErrorSeverity


def render_single_validation(schema: Schema, validation_errors):
    console = Console()
    table = Table(show_header=True, expand=True, box=box.ROUNDED)
    table.add_column("Issue", justify="left")
    for error in validation_errors:
        table.add_row(f"[yellow]{error}[/]")
        table.add_section()
    console.print(table)

def render_validation_output(schema: Schema, validation_errors, config, quiet = False, hide_summary = False, hide_issues = False, show_build = False, show_validation = False, hide_build_summary = False):
    fatal_build_errors = len([x for x in schema.build_errors if x.severity == ErrorSeverity.ERROR])
    non_fatal_build_errors = len(schema.build_errors) - fatal_build_errors
    total_validation_errors = sum(len(v) for k,v in validation_errors.items())
        
    console = Console()

    if quiet:
       hide_summary = True
       hide_issues = True
        
    if show_build:
        table = Table(show_header=True, expand=True, box=box.ROUNDED)
        table.add_column("Build input", justify="left")
        table.add_column("Status", justify="left", width=10)
        for input in [config['coreSchema']] + config['supportingSchemas']:
            errors_for_input = len([b for b in schema.build_errors if b.path == Path(input)])
            table.add_row(f"[bright_blue]{input}[/]", "[green]OK[/]" if errors_for_input == 0 else f"[bright_red]{errors_for_input} ERRORS[/]")
        console.print(table)
    else:
        logging.debug("Hiding build output (use --showbuild to show)")

    if show_validation:
        if fatal_build_errors > 0:
            table = Table(show_header=False, expand=True, box=box.ROUNDED)
            table.add_row("[bright_red]No validation performed due to build failures[/]")
        else:
            table = Table(show_header=True, expand=True, box=box.ROUNDED)
            table.add_column("Validation input", justify="left")
            table.add_column("Status", justify="left", width=10)
            for input in config['instanceDocs']:                
                errors_for_input = len(validation_errors[input])
                table.add_row(f"[bright_blue]{input}[/]", "[green]OK[/]" if errors_for_input == 0 else f"[bright_red]{errors_for_input} ERRORS[/]")
        console.print(table)
    else:
        logging.debug("Hiding validation output (use --showvalidation to show)")

    if hide_issues:
        logging.debug("Hiding issues list (omit --hideissues to show)")
    else:
        if len(schema.build_errors) > 0:
            table = Table(show_header=True, expand=True, box=box.ROUNDED)
            table.add_column("Schema", width=8, justify="left")
            table.add_column("Issue", justify="left")
            table.add_column("Severity", justify="left")
            for build_error in schema.build_errors:
                table.add_row(f"[bright_blue]{build_error.path}[/]", f"[yellow]{build_error.error}[/]", f"[bright_red]ERROR[/]" if build_error.severity == ErrorSeverity.ERROR else "[yellow]WARNING[/]")
            console.print(table)
            
        if total_validation_errors > 0:
            for path, error_list in validation_errors.items():
                if len(error_list) == 0:
                    continue
                table = Table(show_header=False, expand=True, box=box.ROUNDED)
                table.add_row(f"[bright_blue]{path}:[/] [bright_red]{len(error_list)}[/] issues")
                for error in error_list:
                    table.add_section()
                    table.add_row(f"[yellow]{error}[/]")
                console.print(table)

    if hide_summary:
        logging.debug("Hiding stats (omit --hidesummary to show)")
    else:
        table = Table(show_header=False, expand=True, box=box.ROUNDED)
        table.add_column("Field", width=8, justify="left")
        table.add_column("Value", justify="left")
        if not hide_build_summary:
            table.add_row("Schemas", f"[bright_blue]{config['coreSchema']}[/] plus [bright_blue]{len(config['supportingSchemas'])}[/] supporting schemas")
            table.add_row("Build errors", f"{"[green]" if len(schema.build_errors) == 0 else "[bright_red]"}{fatal_build_errors}[/] fatal, {"[green]" if non_fatal_build_errors == 0 else "[yellow]"}{non_fatal_build_errors}[/] non-fatal")
            table.add_section()
        table.add_row("Instance docs validated", f"[bright_blue]{len(config['instanceDocs'])}[/]")
        if (len(config['instanceDocs'])) > 0:
            table.add_row("Validation failures", f"{"[green]" if total_validation_errors == 0 else "[bright_red]"}{total_validation_errors}[/]")
        table.add_section()
        if fatal_build_errors > 0:
            table.add_row("Outcome", "[bright_red]FAILED BUILD[/]")
        elif total_validation_errors > 0:
            table.add_row("Outcome", "[bright_red]FAILED VALIDATION[/]")
        else:
            table.add_row("Outcome", "[green]OK[/]")
        console.print(table)


from pathlib import Path

from forgeschema import Schema


def test_path_expension():
    s = Schema(Path('test/data/test.x'), [Path('test/data/',)], ['x'])
    assert s.core_schema_path == Path('test/data/test.x')
    assert s.supporting_schemas_paths == [Path('test/data/test.x'), Path('test/data/inner1/test2.x'), Path('test/data/inner1/inner2/test3.x')]

    s = Schema(Path('test/data/test.x'), [Path('test/data/',)], ['x', 'y'])
    assert s.core_schema_path == Path('test/data/test.x')
    expected = [Path('test/data/test.x'), Path('test/data/test.y'),
                 Path('test/data/inner1/test2.y'), Path('test/data/inner1/test2.x'),
                 Path('test/data/inner1/inner2/test3.x'), Path('test/data/inner1/inner2/test3.y')]
    diff = set(s.supporting_schemas_paths) ^ set(expected)
    assert not diff



from pathlib import Path
from typing import List

from xmlschema import XMLSchemaValidationError, XMLSchemaParseError
from urllib.error import URLError
from xml.etree.ElementTree import ParseError
from pytest import mark

from forgeschema import XSDSchema
from forgeschema.types import ErrorSeverity



def test_xsd():
    s = XSDSchema(Path('test/xsd/set1/good.xsd'), [])
    s.build()
    assert s.built_ok

@mark.parametrize("instance_path, expected_exception", [
    (Path('test/xsd/set1/bad_namespace.xml'), XMLSchemaValidationError),
    (Path('test/xsd/set1/bad_not_compliant.xml'), XMLSchemaValidationError),
    (Path('test/xsd/set1/bad_syntax.xml'), ParseError),
])
def test_simple_fails_validation(instance_path: Path, expected_exception):
    s = XSDSchema(Path('test/xsd/set1/good.xsd'), [])
    s.build()
    errors = s.validate(instance_path)
    assert len(errors) == 1
    assert isinstance(errors[0], expected_exception)


@mark.parametrize("core_schema, expected_exception", [
    (Path('thisfiledoesntexist'), URLError),
    (Path('test/xsd/set1/bad_syntax.xml'), ParseError)
])
def test_bad(core_schema: Path, expected_exception):
    s = XSDSchema(core_schema, [])
    s.build()
    assert not s.built_ok
    assert len(s.build_errors) == 1
    b = s.build_errors[0]
    assert isinstance(b.error, expected_exception) 


def test_bad_supporting():
    s = XSDSchema(Path('thisfiledoesntexist'), [Path('nordoesthisone.xsd')])
    s.build()
    assert not s.built_ok
    assert len(s.build_errors) == 2
    assert isinstance(s.build_errors[0].error, URLError) 
    assert isinstance(s.build_errors[1].error, URLError) 


@mark.parametrize("core_schema, supporting_schemas, instance_files",[
    (Path('test/xsd/set1/good.xsd'),[],[Path('test/xsd/set1/good.xml')]),
    (Path('test/xsd/set2/good_with_import.xsd'),
         [Path('test/xsd/set2/good_imported.xsd')],
         [Path('test/xsd/set2/good.xml')])
])
def test_matches_xsd(core_schema : Path, supporting_schemas : List[Path], instance_files : List[Path]):
    s = XSDSchema(core_schema, supporting_schemas)
    s.build()
    assert s.built_ok
    for instance in instance_files:
        errors = s.validate(instance)
        assert len(errors) == 0

@mark.parametrize("core_schema, expected_error",[
    (Path('test/xsd/bad_xsds/missing_element_name.xsd'),XMLSchemaParseError),
    (Path('test/xsd/bad_xsds/complex_content.xsd'),XMLSchemaParseError)
])
def test_bad_xsds(core_schema : Path, expected_error):
    s = XSDSchema(Path(core_schema), [])
    s.build()
    assert not s.built_ok
    assert len(s.build_errors) == 1
    assert isinstance(s.build_errors[0].error, expected_error)


from typing import List, Generator, Tuple
from pathlib import Path

from xmlschema import XMLSchema, XMLSchemaValidatorError
from xml.etree.ElementTree import ParseError

from .schema import Schema
from .types import BuildError, ErrorSeverity

class XSDSchemaResource:
    path : Path
    schema : XMLSchema
    namespace : str
    
    def __init__(self, path, build_errors: List[BuildError]):
        self.path = path
        self.schema = XMLSchema(str(path), validation='skip')
        self.namespace = self.schema.target_namespace
        # For some reason we get spurious UPA errors in the supporting schemas
        # so these are filtered out. A real UPA error will be caught when the
        # schema set if built with "strict" later on.
        build_errors += [
            BuildError(path, ErrorSeverity.ERROR, error)
            for error in self.schema.all_errors
            if not error.message.startswith("Unique Particle Attribution violation")
        ]


class XSDSchema(Schema):
    def __init__(self, core_schema : Path, supporting_schemas : List[Path]):
        super().__init__(core_schema, supporting_schemas, ["xsd"])
    
    def build(self) -> bool:
        self.build_errors.clear()
        self.built_ok = False

        self.supporting_schemas = []
        for schema_path in self.supporting_schemas_paths:
            try:
                self.supporting_schemas.append(XSDSchemaResource(schema_path, self.build_errors))
            except Exception as ex:
                self.build_errors.append(BuildError(schema_path, ErrorSeverity.ERROR, ex))
        
        schema_locations = [(r.namespace, str(r.path.resolve())) for r in self.supporting_schemas]

        try:
            self.core_schema = XMLSchema(str(self.core_schema_path), locations = schema_locations, validation='strict')
        except Exception as ex:
            self.build_errors.append(BuildError(self.core_schema_path, ErrorSeverity.ERROR, ex))
            return False

        self.built_ok = True
        return True

    def validate_string(self, string:str) -> List[Exception]:
        try:
            return list(self.core_schema.iter_errors(string))
        except ParseError as ex:
            return [ex]

    def validate(self, instance_doc: Path) -> List[Exception]:
        if not instance_doc.exists():
            return [FileNotFoundError(instance_doc)]
        try:
            errors = list(self.core_schema.iter_errors(str(instance_doc)))
        except Exception as ex:
            errors = [ex]
        return errors



from typing import List
from pathlib import Path

from json import loads, JSONDecodeError
from xml.etree.ElementTree import parse, ParseError

import logging

def expand_path(path: Path, globs:List[str] | str, recursive: bool = True, case_sensitive = False):
    if not path.exists():
        return [path]
    if not path.is_dir():
        return [path]
    if recursive:
        if isinstance(globs, str):
            return [f for f in path.rglob(globs, case_sensitive=case_sensitive)]
        else:
            return [f for glob in globs for f in path.rglob(glob, case_sensitive=case_sensitive)]
    else:
        if isinstance(globs, str):
            return [f for f in path.glob(globs, case_sensitive=case_sensitive)]
        else:
            return [f for glob in globs for f in path.glob(glob, case_sensitive=case_sensitive)]

def guess_encoding (file: Path):
    logging.debug(f"Attempting to guess encoding of {file}")
    if file.suffix.lower() == ".json":
        logging.debug(f"Guessing json from extension")
        return "json"
    if file.suffix.lower() in [".xml", ".xsd"]:
        logging.debug(f"Guessing XML from extension")
        return "xml"

    if not file.exists():
        logging.warning(f"Couldn't open {file} to check contents for encoding")
        return None
    
    try:
        j = loads(file.read_text())
        logging.info(f"JSON encoding detected contents of {file}")
        return "json"
    except JSONDecodeError as ex:
        logging.debug("Decoding contents as JSON failed")

    try:
        et = parse(str(file))
        logging.info(f"XML encoding detected from contents of {file}")
        return "xml"
    except ParseError as ex:
        logging.debug("Decoding contents as XML failed")

    return None


from pathlib import Path
from os import getenv
from pytest import raises

from forgeschema.helpers import expand_path



def test_path_expansion_notexisting():
    assert expand_path(Path('not_a_file'), "*.*") == [Path('not_a_file')]

def test_path_expansion_single_file():
    single_file = Path('test/data/test.x')
    r = expand_path(single_file, "blah")
    assert r == [single_file]

def test_path_expansion_dir_non_recursive():
    root_path = Path('test/data/')
    r = expand_path(root_path, "*.x", recursive=False)
    assert r == [Path('test/data/test.x')]
    r = expand_path(root_path, "*.y", recursive=False)
    assert r == [Path('test/data/test.y')]
    r = expand_path(root_path, "*.*", recursive=False)
    assert r == [Path('test/data/test.x'), Path('test/data/test.y')]
    
def test_path_expansion_dir_recursive():
    root_path = Path('test/data/')
    r = expand_path(root_path, "*.x", recursive=True)
    assert r == [Path('test/data/test.x'), Path('test/data/inner1/test2.x'), Path('test/data/inner1/inner2/test3.x')]
    r = expand_path(root_path, "*.y", recursive=True)
    assert r == [Path('test/data/test.y'), Path('test/data/inner1/test2.y'), Path('test/data/inner1/inner2/test3.y')]

def test_path_expansion_multiglob():
    root_path = Path('test/data/')
    r = expand_path(root_path, ['*.x', '*.y'], recursive=False)
    assert r == [Path('test/data/test.x'), Path('test/data/test.y')]
    r = expand_path(root_path, ['*.x', '*.y'], recursive=True)
    expected = [Path('test/data/test.x'), Path('test/data/test.y'),
                 Path('test/data/inner1/test2.y'), Path('test/data/inner1/test2.x'),
                 Path('test/data/inner1/inner2/test3.x'), Path('test/data/inner1/inner2/test3.y')]
    diff = set(r) ^ set(expected)
    assert not diff





import logging
import json
from argparse import ArgumentParser
from pathlib import Path
from xml.etree import ElementTree

from rich.table import Table
from rich import box
from rich.console import Console

from ..schema import Schema
from ..jsonschema import JSONSchema
from ..xsdschema import XSDSchema
from ..types import ErrorSeverity
from ..helpers import guess_encoding, expand_path
from ..render import render_validation_output


ERR_FILE_NOT_FOUND          = -1
ERR_UNSUPPORTED_ENCODING    = -2
ERR_UNKNOWN_ENCODING        = -3
ERR_BAD_CONFIG              = -4

EXIT_OK                     = 0
EXIT_BUILD_ERRORS           = 1
EXIT_VALIDATION_ERRORS      = 2

supported_encodings = {
    'json' : {
        'schema_extensions' : ['.schema.json'],
        'instance_extensions' : ['.json']
    },
    'xml' : {
        'schema_extensions' : ['.xsd'],
        'instance_extensions' : ['.xml']
    }
}

def check():
    parser = ArgumentParser()

    parser = ArgumentParser(description="Build a schema set and use it to validate zero or more instance documents. Returns 1 if the schema build fails, 2 if any of the instance documents fail to validate, or 0 otherwise. If JSON output is selected via the -j switch, it is always 0")
    parser.add_argument("-e", "--encoding", help=f"Set the encoding of the schema / instance documents. One of [{'|'.join(list(supported_encodings.keys()))}]")
    arg_group = parser.add_mutually_exclusive_group(required=True)
    arg_group.add_argument("-c", "--config", help="Specifies the location of a JSON config file")
    arg_group.add_argument("-s", "--coreschema", help="Specifies the location of the core JSON schema file")
    parser.add_argument("-u", "--supportingschema", action="append", help="Specifies the location of any supporting schemas required. If a directory is specified, testjson will search and add any .schema.json files recursively within the directory")
    parser.add_argument("-i", "--instancedoc", action="append", help="Instance XML document to validate against the schema. If a directory is specified, xmltest will search and add any XML files recursively within the directory")
    parser.add_argument("-j", "--jsonoutput", action="store_true", help="Output JSON instead of text. Return code will always be zero")
    parser.add_argument("-v", "--verbose", action="count", help="Verbose. Can be specified multiple times to get more detailed output")
    parser.add_argument("--showbuild", action="store_true", help="Print build output to stdout")
    parser.add_argument("--showvalidation", action="store_true", help="Print validation output to stdout")
    parser.add_argument("--hideissues", action="store_true", help="Print list of build and validation issues to stdout")
    parser.add_argument("--hidesummary", action="store_true", help="Hide summary output on stdout")
    parser.add_argument("-q", "--quiet", action="store_true", help="Hide all output (equivalent to --hideissues plus --hidesummary)")
    pargs = parser.parse_args()

    if pargs.verbose is None or pargs.verbose == 1:
        logging.basicConfig(level=logging.ERROR)
    elif pargs.verbose == 2:
        logging.basicConfig(level=logging.WARNING)
    elif pargs.verbose == 3:
        logging.basicConfig(level=logging.INFO)
    elif pargs.verbose > 3:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.ERROR)
    
    logging.debug(f"Called with arguments {pargs}")

    if pargs.config is not None:
        config_file = Path(pargs.config)
        if not config_file.exists():
            logging.error(f"Cannot find config file {pargs.config}")
            exit(ERR_FILE_NOT_FOUND)
        config = json.loads(config_file.read_text())
        if 'coreSchema' not in config.keys():
            logging.error('Config file does not have a "coreSchema" key')
            exit(ERR_BAD_CONFIG)
        if 'supportingSchemas' not in config.keys():
            logging.warning('Config file does not have a "supportingSchemas" key, assuming empty')
            config['supportingSchemas'] = []
        if 'instanceDocs' not in config.keys():
            logging.warning('Config file does not have a "instanceDocs" key, assuming empty')
            config['instanceDocs'] = []
    else:
        config = {
            'coreSchema' : pargs.coreschema,
            'supportingSchemas' : [s for s in pargs.supportingschema] if pargs.supportingschema is not None else [],
            'instanceDocs' : [i for i in pargs.instancedoc] if pargs.instancedoc is not None else [],
        }

    if pargs.encoding is None:
        if not (Path(config['coreSchema']).exists()):
            logging.error(f"Could not load core schema {config['coreSchema']}")
            exit(ERR_FILE_NOT_FOUND)
        logging.debug("No encoding specified, attempting to guess encoding from core schema...")
        guessed_encoding = guess_encoding(Path(config['coreSchema']))
        if guessed_encoding is None:
            logging.error(f"Couldn't guess encoding from {config['coreSchema']}")
            exit(ERR_UNKNOWN_ENCODING)
        else:
            pargs.encoding = guessed_encoding

    if pargs.encoding not in list(supported_encodings.keys()):
        logging.error(f"Unsupported encoding '{pargs.encoding}' specified. Supported encodings are [{'|'.join(list(supported_encodings.keys()))}]")
        exit(ERR_UNSUPPORTED_ENCODING)
    if pargs.encoding == 'xml':
        logging.debug("Creating XML schema")
        schema = XSDSchema(Path(config['coreSchema']), [Path(x) for x in config['supportingSchemas']])
    if pargs.encoding == 'json':
        logging.debug("Creating JSON schema")
        schema = JSONSchema(Path(config['coreSchema']), [Path(x) for x in config['supportingSchemas']])


    logging.debug("Expanding supporting schemas...")
    config['supportingSchemas'] = [expanded_path for path in config['supportingSchemas'] for expanded_path in expand_path(Path(path), [f'*{x}' for x in supported_encodings[pargs.encoding]['schema_extensions']])]
    if Path(config['coreSchema']) in config['supportingSchemas']:
        config['supportingSchemas'].remove(Path(config['coreSchema']))

    logging.debug("Expanding instance docs...")
    config['instanceDocs'] = [expanded_path for path in config['instanceDocs'] for expanded_path in expand_path(Path(path), [f'*{x}' for x in supported_encodings[pargs.encoding]['instance_extensions']])]

    logging.debug(f"Config: {config}")

    schema.build()

    if schema.built_ok:
        logging.info('Schema built OK')
        validation_errors = {f : schema.validate(Path(f)) for f in config['instanceDocs']}
    else:
        logging.warning('Errors building schema')
        validation_errors = {}

    if pargs.jsonoutput:
        logging.debug("Giving output in JSON format")
        output = {
            'input' : config,
            'build_errors' : [be.toJSON() for be in schema.build_errors],
            'validation_errors' : validation_errors
        }
        print(json.dumps(output, default=str))
        exit(EXIT_OK)

    render_validation_output(schema, validation_errors, config, pargs.quiet, pargs.hidesummary, pargs.hideissues, pargs.showbuild, pargs.showvalidation)

    # fatal_build_errors = len([x for x in schema.build_errors if x.severity == ErrorSeverity.ERROR])
    # non_fatal_build_errors = len(schema.build_errors) - fatal_build_errors
    total_validation_errors = sum(len(v) for k,v in validation_errors.items())
        
    # console = Console()

    # if pargs.quiet:
    #     pargs.hidesummary = True
    #     pargs.hideissues = True
        
    # if pargs.showbuild:
    #     table = Table(show_header=True, expand=True, box=box.ROUNDED)
    #     table.add_column("Build input", justify="left")
    #     table.add_column("Status", justify="left", width=10)
    #     for input in [config['coreSchema']] + config['supportingSchemas']:
    #         errors_for_input = len([b for b in schema.build_errors if b.path == Path(input)])
    #         table.add_row(f"[bright_blue]{input}[/]", "[green]OK[/]" if errors_for_input == 0 else f"[bright_red]{errors_for_input} ERRORS[/]")
    #     console.print(table)
    # else:
    #     logging.debug("Hiding build output (use --showbuild to show)")

    # if pargs.showvalidation:
    #     if fatal_build_errors > 0:
    #         table = Table(show_header=False, expand=True, box=box.ROUNDED)
    #         table.add_row("[bright_red]No validation performed due to build failures[/]")
    #     else:
    #         table = Table(show_header=True, expand=True, box=box.ROUNDED)
    #         table.add_column("Validation input", justify="left")
    #         table.add_column("Status", justify="left", width=10)
    #         for input in config['instanceDocs']:                
    #             errors_for_input = len(validation_errors[input])
    #             table.add_row(f"[bright_blue]{input}[/]", "[green]OK[/]" if errors_for_input == 0 else f"[bright_red]{errors_for_input} ERRORS[/]")
    #     console.print(table)
    # else:
    #     logging.debug("Hiding validation output (use --showvalidation to show)")

    # if pargs.hideissues:
    #     logging.debug("Hiding issues list (omit --hideissues to show)")
    # else:
    #     if len(schema.build_errors) > 0:
    #         table = Table(show_header=True, expand=True, box=box.ROUNDED)
    #         table.add_column("Schema", width=8, justify="left")
    #         table.add_column("Issue", justify="left")
    #         table.add_column("Severity", justify="left")
    #         for build_error in schema.build_errors:
    #             table.add_row(f"[bright_blue]{build_error.path}[/]", f"[yellow]{build_error.error}[/]", f"[bright_red]ERROR[/]" if build_error.severity == ErrorSeverity.ERROR else "[yellow]WARNING[/]")
    #         console.print(table)
            
    #     if total_validation_errors > 0:
    #         table = Table(show_header=True, expand=True, box=box.ROUNDED)
    #         table.add_column("Document", width=8, justify="left")
    #         table.add_column("Issue", justify="left")
    #         for path, error_list in validation_errors.items():
    #             for error in error_list:
    #                 table.add_row(f"[bright_blue]{path}", f"[yellow]{error}[/]")
    #                 table.add_section()
    #         console.print(table)

    # if pargs.hidesummary:
    #     logging.debug("Hiding stats (omit --hidesummary to show)")
    # else:
    #     table = Table(show_header=False, expand=True, box=box.ROUNDED)
    #     table.add_column("Field", width=8, justify="left")
    #     table.add_column("Value", justify="left")
    #     table.add_row("Schemas", f"[bright_blue]{config['coreSchema']}[/] plus [bright_blue]{len(config['supportingSchemas'])}[/] supporting schemas")
    #     table.add_row("Build errors", f"{"[green]" if len(schema.build_errors) == 0 else "[bright_red]"}{fatal_build_errors}[/] fatal, {"[green]" if non_fatal_build_errors == 0 else "[yellow]"}{non_fatal_build_errors}[/] non-fatal")
    #     table.add_section()
    #     table.add_row("Instance docs validated", f"[bright_blue]{len(config['instanceDocs'])}[/]")
    #     if (len(config['instanceDocs'])) > 0:
    #         table.add_row("Validation failures", f"{"[green]" if total_validation_errors == 0 else "[bright_red]"}{total_validation_errors}[/]")
    #     table.add_section()
    #     if fatal_build_errors > 0:
    #         table.add_row("Outcome", "[bright_red]FAILED BUILD[/]")
    #     elif total_validation_errors > 0:
    #         table.add_row("Outcome", "[bright_red]FAILED VALIDATION[/]")
    #     else:
    #         table.add_row("Outcome", "[green]OK[/]")
    #     console.print(table)

    if not schema.built_ok:
        logging.warning(f"Schema had {len(schema.build_errors) } build errors. Exiting with code {EXIT_BUILD_ERRORS}")
        exit(EXIT_BUILD_ERRORS)
    if total_validation_errors > 0:
        logging.warning(f"Schema built OK but there were {total_validation_errors} validation errors in instance documents. Exiting with code {EXIT_VALIDATION_ERRORS}")
        exit(EXIT_VALIDATION_ERRORS)

    exit(EXIT_OK)


from typing import List, Generator, Tuple
from pathlib import Path
from json import loads, JSONDecodeError
import logging

from referencing import Registry
from referencing.exceptions import Unresolvable
from jsonschema import Draft202012Validator
from jsonschema.exceptions import ValidationError

from .schema import Schema
from .types import BuildError, ErrorSeverity

class MissingKeyException (Exception):
    pass

class JSONSchemaResource:
    path : Path
    raw : str
    json : dict
    id : str
    spec : str

    def __init__(self, path, build_errors : List[BuildError]):
        self.path = path
        self.raw = path.read_text()
        self.json = loads(self.raw)
        if "$id" in self.json.keys():
            self.id = self.json['$id']
        else:
            raise MissingKeyException("Schema has no $id key")
        if '$schema' in self.json.keys():
            self.spec = self.json['$schema']
        else:
            build_errors.append(BuildError(path, ErrorSeverity.WARNING, MissingKeyException(f"Schema {self.id} has no specification ('$schema') key")))
            self.spec = "https://json-schema.org/draft/2020-12/schema"
            self.json['$schema'] = "https://json-schema.org/draft/2020-12/schema"

class JSONSchema(Schema):
    def __init__(self, core_schema : Path, supporting_schemas : List[Path], expand_errors : bool = True):
        super().__init__(core_schema, supporting_schemas, ["schema.json"])
        self.expand_errors = expand_errors

    def build(self) -> bool:
        self.build_errors.clear()
        self.built_ok = False

        try:
            self.core_schema = JSONSchemaResource(self.core_schema_path, self.build_errors)
        except Exception as ex:
            self.build_errors.append(BuildError(self.core_schema_path, ErrorSeverity.ERROR, ex))
            self.core_schema = None

        self.supporting_schemas = []
        for path in self.supporting_schemas_paths:
            try:
                self.supporting_schemas.append(JSONSchemaResource(path, self.build_errors))
            except Exception as ex:
                self.build_errors.append(BuildError(path, ErrorSeverity.ERROR, ex))
        
        if self.core_schema is None:
            return False
        
        try:
            self.registry = Registry().with_contents([(r.id, r.json) for r in self.supporting_schemas] + [(self.core_schema.id, self.core_schema.json)])
            self.validator = Draft202012Validator(self.core_schema.json, registry=self.registry)
            Draft202012Validator.check_schema(self.core_schema.json)
        except Exception as ex:
            self.build_errors.append(BuildError(None, ErrorSeverity.ERROR, ex))
            return False
        
        ref_errors = self._check_all_refs()
        if len(ref_errors) > 0:
            self.build_errors += ref_errors
            return False
        
        self.built_ok = True
        return True

    def _check_all_refs(self) -> List[Exception]:
        errors = self._check_refs(self.core_schema)
        for other_schema in self.supporting_schemas:
            errors += self._check_refs(other_schema)
        return errors
    
    def _check_refs(self, schema) -> List[Exception]:
        refs = list(self._recursively_find_refs(schema.json))
        resource = self.registry.get(schema.id)
        resolver = self.registry.resolver_with_root(resource)
        errors = []
        for key, ref in refs:
            try:
                resolver.lookup(ref)
            except Unresolvable as ex:
                errors.append(BuildError(schema.path, ErrorSeverity.ERROR, ex))
        return errors
    
    def _recursively_find_refs(self, j) -> Generator[Tuple[str, str], None, None]:
        if isinstance(j, dict):            
            for k,v in j.items():
                if k == "$ref":
                    yield k,v
                else:
                    yield from self._recursively_find_refs(v)
        elif isinstance(j, list):
            for i in j:
                yield from self._recursively_find_refs(i)
        else:
            return

    def validate_string(self, string: str) -> List[Exception]:
        errors = []
        try:
            instance_json = loads(string)
        except JSONDecodeError as ex:
            return [ex]        
        for error in self.validator.iter_errors(instance_json):
            errors += self.expandError(error)
        return errors

    def validate(self, instance_doc: Path) -> List[Exception]:
        try:
            return self.validate_string(instance_doc.read_text())
        except (PermissionError, FileNotFoundError) as ex:
            return [ex]

    def expandError(self, ex : Exception):
        if not self.expand_errors:
            return ex
        if not isinstance(ex, ValidationError):
            logging.debug("Not a Validation Error " + str(type(ex)))
            return ex
        if ex.schema_path[-1] == "oneOf":
            logging.debug(ex.absolute_path)
            logging.debug("Looks like a oneOf problem")
            logging.debug("Valid options are...")
            for thing in ex.schema['oneOf']:
                logging.debug ("Maybe " + str(thing))
                if "$ref" in thing:
                    resolved = self.registry.resolver().lookup(thing['$ref'])
                    candidate_def = resolved.contents
                else:
                    candidate_def = thing

                if ("required" in candidate_def
                    and len(candidate_def["required"]) > 0
                    and candidate_def["required"][0] in ex.instance
                    ):
                    candidate_field = candidate_def["required"][0]
                    candidate_ref = candidate_def["properties"][candidate_field]
                    if (candidate_field == '@xsi:type'
                        and 'properties' in candidate_def
                        and '@xsi:type' in candidate_def['properties']):
                        expected_type = candidate_def['properties']['@xsi:type']['enum'][0]
                        if '@xsi:type' in ex.instance:
                            if ex.instance['@xsi:type'] == expected_type:
                                candidate_ref = thing
                                test_instance = ex.instance
                            else:
                                logging.debug("xsi type doesn't match, moving on")
                                continue
                        else:
                            logging.debug ("No xsi type in instance, moving on")
                            continue
                    else:
                        logging.debug("Doesn't look like xsi:type, trying ref directly")
                        #TODO need to determine correct source document in order for this to 
                        # work e.g. in TrafficPolicy
                        likely_ns = self.validator.schema["$id"]
                        if candidate_field.startswith("tp:"):
                            likely_ns = 'ts_103120_TrafficPolicy_2022_07'
                            # quick hack
                            # TODO remove this
                        candidate_ref['$ref'] = likely_ns + candidate_ref['$ref']
                        try:
                            resolved = self.registry.resolver().lookup(candidate_ref['$ref'])
                        except Exception as inner_ex:
                            logging.debug(f"Failed resolving {candidate_ref} ({inner_ex}), carrying on")
                            continue
                        if not resolved:
                            logging.debug(f"Could resolve {candidate_ref}, carrying on")
                            continue
                        test_instance = ex.instance[candidate_field]
                else:
                    logging.debug("Doesn't look like it matches this oneOf option, moving on")
                    continue
                
                logging.debug ("Having a go at matching:")
                logging.debug(candidate_ref)
                logging.debug(test_instance)
                fragment_validator = Draft202012Validator(candidate_ref, registry=self.registry)
                new_errors = list(fragment_validator.iter_errors(test_instance))
                for error in new_errors:
                    expanded_errors = self.expandError(error)
                    if len(expanded_errors) == 0:
                        return ex
                    else:
                        for expanded_error in expanded_errors:
                            expanded_error.path = ex.path + error.path
                        return expanded_errors
            logging.debug("No valid options, returning self")
            return [ex]
        else:
            logging.debug("Some other kind of error")
            return [ex]